@@ -13,6 +13,7 @@ from redis.commands.json.path import Path
 from redis.commands.search import Search
 from redis.commands.search.field import (
     GeoField,
+    GeoShapeField,
     NumericField,
     TagField,
     TextField,
@@ -2226,7 +2227,7 @@ def test_withsuffixtrie(client: redis.Redis):
         assert "WITHSUFFIXTRIE" not in info["attributes"][0]
         assert client.ft().dropindex("idx")
 
-        # create withsuffixtrie index (text fiels)
+        # create withsuffixtrie index (text fields)
         assert client.ft().create_index((TextField("t", withsuffixtrie=True)))
         waitForIndex(client, getattr(client.ft(), "index_name", "idx"))
         info = client.ft().info()
@@ -2243,7 +2244,7 @@ def test_withsuffixtrie(client: redis.Redis):
         assert "WITHSUFFIXTRIE" not in info["attributes"][0]["flags"]
         assert client.ft().dropindex("idx")
 
-        # create withsuffixtrie index (text fiels)
+        # create withsuffixtrie index (text fields)
         assert client.ft().create_index((TextField("t", withsuffixtrie=True)))
         waitForIndex(client, getattr(client.ft(), "index_name", "idx"))
         info = client.ft().info()
@@ -2266,3 +2267,20 @@ def test_query_timeout(r: redis.Redis):
     q2 = Query("foo").timeout("not_a_number")
     with pytest.raises(redis.ResponseError):
         r.ft().search(q2)
+
+
+@pytest.mark.redismod
+def test_geoshape(client: redis.Redis):
+    client.ft().create_index((GeoShapeField("geom", GeoShapeField.FLAT)))
+    waitForIndex(client, getattr(client.ft(), "index_name", "idx"))
+    client.hset("small", "geom", "POLYGON((1 1, 1 100, 100 100, 100 1, 1 1))")
+    client.hset("large", "geom", "POLYGON((1 1, 1 200, 200 200, 200 1, 1 1))")
+    q1 = Query("@geom:[WITHIN $poly]").dialect(3)
+    qp1 = {"poly": "POLYGON((0 0, 0 150, 150 150, 150 0, 0 0))"}
+    q2 = Query("@geom:[CONTAINS $poly]").dialect(3)
+    qp2 = {"poly": "POLYGON((2 2, 2 50, 50 50, 50 2, 2 2))"}
+    result = client.ft().search(q1, query_params=qp1)
+    assert len(result.docs) == 1
+    assert result.docs[0]["id"] == "small"
+    result = client.ft().search(q2, query_params=qp2)
+    assert len(result.docs) == 2
